/***
 * GASP: a gaming services platform for mobile multiplayer games.
 * MooDS: a communication protocol used in GASP.
 * GASP Demonstrator: a gasp demonstration game.
 *
 * Copyright (C) 2004 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Franck Caron, Romain Pellerin
 */
/*
 * Created on 2 december 2005
 *
 * @author CARON Franck - CNAM CEDRIC
 * @author PELLERIN Romain - CNAM CEDRIC
 * 
 * 
 * Research project: CNAM-CEDRIC - INT - Filao
 */
package app2;


import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.util.Hashtable;
import java.util.Enumeration;
import app2.NewPlayer;
import app2.UpdatePlayer;
import app2.PingPlayer;

// specifical ping case
import app2.client.NetworkCom;
import app2.client.game.SimpleGameCanvas;


public class CustomTypes{ 
	 public void encodeNewPlayer(DataOutputStream dos,NewPlayer customType) throws Exception { 
		 dos.writeShort(customType.getId());
		 dos.writeInt(customType.getX());
		 dos.writeInt(customType.getY());
		 dos.writeUTF(customType.getUsername());
	 }

	 public void encodeUpdatePlayer(DataOutputStream dos,UpdatePlayer customType) throws Exception { 
		 dos.writeShort(customType.getId());
		 dos.writeInt(customType.getX());
		 dos.writeInt(customType.getY());
		 dos.writeInt(customType.getSpeedX());
		 dos.writeInt(customType.getSpeedY());
	 }

	 public void encodePingPlayer(DataOutputStream dos,PingPlayer customType) throws Exception { 
		 dos.writeShort(customType.getId());
		 dos.writeInt(customType.getPingNum());
		 dos.writeInt(customType.getSenderAveragePing());
		 dos.writeInt(customType.getDesynchronizationWithSender());
	 }
	 
	 public void encodeData(Hashtable h, DataOutputStream dos) throws Exception {
		 dos.writeByte(h.size());
		 int nbSend = h.size();
		 for(int i=0; i<nbSend; i++){
			 Object o  = (Object) h.get(""+i);
			 if (o.getClass().getName().equals("app2.NewPlayer")){
				 dos.writeByte(6);
				 encodeNewPlayer(dos,(NewPlayer) o);
			 }
			 if (o.getClass().getName().equals("app2.UpdatePlayer")){
				 dos.writeByte(7);
				 encodeUpdatePlayer(dos,(UpdatePlayer) o);
			 }
			 if (o.getClass().getName().equals("app2.PingPlayer")){
				 dos.writeByte(8);
				 encodePingPlayer(dos,(PingPlayer) o);
			 }
		}
	}
	 public NewPlayer decodeNewPlayer(DataInputStream dis) throws Exception { 
		 NewPlayer customType = new NewPlayer();
		 customType.setId(dis.readShort()); 
		 customType.setX(dis.readInt()); 
		 customType.setY(dis.readInt()); 
		 customType.setUsername(dis.readUTF()); 

		 return customType;
	 }

	 public UpdatePlayer decodeUpdatePlayer(DataInputStream dis) throws Exception { 
		 UpdatePlayer customType = new UpdatePlayer();
		 customType.setId(dis.readShort()); 
		 customType.setX(dis.readInt()); 
		 customType.setY(dis.readInt()); 
		 customType.setSpeedX(dis.readInt()); 
		 customType.setSpeedY(dis.readInt()); 

		 return customType;
	 }
	 
	 public PingPlayer decodePingPlayer(DataInputStream dis) throws Exception { 
		 PingPlayer customType = new PingPlayer();
		 customType.setId(dis.readShort()); 
		 customType.setPingNum(dis.readInt());
		 customType.setSenderAveragePing(dis.readInt()); 
		 customType.setDesynchronizationWithSender(dis.readInt()); 
		 return customType;
	 }

	 public Hashtable decodeData(DataInputStream dis) throws Exception {
		 Hashtable h = new Hashtable();
		 byte nb = dis.readByte();
		 int nbObj = nb;
		 while(nb!=0){
			 String key = ""+(nbObj - nb);
			 byte type = dis.readByte();
			 if (type == 6){
				 NewPlayer customType = decodeNewPlayer(dis);
				 h.put(key,customType);
			 }
			 else if (type == 7){
				 UpdatePlayer customType = decodeUpdatePlayer(dis);
				 h.put(key,customType);
			 }
			 else if (type == 8){
				 PingPlayer customType = decodePingPlayer(dis);
				 h.put(key,customType);
			 }
			 nb--;
		}
		 return h;
	}
}