/***
 * GASP: a gaming services platform for mobile multiplayer games.
 * Copyright (C) 2004 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Romain Pellerin
 */
//#include parameters.h
////#define DEMO
////#define VERSION_SOLO
////#define FRANCK
////#define LASTPING
////#define RESYNC

//#endinclude
package app2.client;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Vector;

import javax.microedition.io.Connector;
import javax.microedition.io.HttpConnection;

import app2.CustomTypes;



/**
 * A game client logic running with the MGP must implement this class, MGPClient.
 * 
 * @author PELLERIN Romain (pellerin@cnam.fr) - MEGA Project  
 */
public abstract class GASPClient {
    
    private static String GPServletContainerUrl = "http://Rom1.infraworlds.int/GASP/servlet/";
	protected CustomTypes customTypes;

	
    abstract public void initCustomTypes();
	
    
	/**
	 * This method is called when the user use the application for the first time, e.g
	 * when the Actor ID is not present on the client.
	 * Returns the ActorID if authentification informations are valids.
	 *
	 * @param applicationID the ID of the application
	 * @param username the username wanted by the user
	 * @param password the password wanted by the user
	 * @return the Actor ID
	 */
    protected int firstLogin(int applicationID,String username, String password) {
    	synchronized(monitor){
	    	String formData ="uid="+username.charAt(4) /** for test replace by a test uid number */
							+"&appID="+applicationID
							+"&username="+username
							+"&password="+password;
			try {
			    dis = servletRequest("FirstLogin", formData);
				return Integer.parseInt(dis.readShort()+""); // ActorID
			} 
			catch (IOException e) {
				e.printStackTrace();
			}
			finally { closeAllConnection();}
			return 0;
    	}
	}

	
	/**
	 * Provide the login on the Gaming Platform, require the correct ActorID, the username
	 * and the password, the version of MGIF is optional.
	 * 
	 * @param aID the Actor ID of the user stored on the mobile
	 * @param username the username choosed at the first login
	 * @param password the password chooset at the first login
	 * @param version MGIF compliant version
	 * 
	 * @return the Session ID of the user
	 */
	protected int login(int aID, String username, String password, String version) {
		synchronized(monitor){
			String formData = "aID="+aID
							+"&username="+username
							+"&password="+password
							+"&version="+version;
			try {
			    dis = servletRequest("Login", formData);
				return Integer.parseInt(dis.readShort()+"");
			} catch (Exception e) {
				e.printStackTrace();
			}
			finally { closeAllConnection();}
			return 0;
		}
	}
		
	private void testMonitorLiberation(){
		synchronized (monitor){
			System.out.println("Monitor libere");
		}
	}
	
	
	/**
	 * Returns a stream of game sessions currently available on the Gaming Platform.
	 * 
	 * @param sID the Session ID of the user
	 * @return the stream format (applicationInstanceID, (actorID,username)*)*
	 */
	protected Vector getApplicationInstances(int sID) {
		System.out.println("getApplicationInstances");
		synchronized(monitor){
			String formData = "sID="+sID;
	 
			try {
			    dis = servletRequest("Lobby", formData);
				return formLobbyDatas(dis);
			} 
			catch (Exception e) {
				e.printStackTrace();
			}
			finally { closeAllConnection();}
			
		}
		testMonitorLiberation();
		return null;	
	}

	

		
	/**
	 * Form the datas contained on the DataInputStream of the lobby servlet response,
	 * returns a vector of ApplicationInstanceInfos objects.
	 * 
	 * @return the vector of ApplicationInstanceInfos objects
	 * @throws IOException
	 */
	private Vector formLobbyDatas(DataInputStream dis) throws IOException {	    
		Vector appInstances = new Vector();
			
		// read the number of applications instances	
		short aIIDs = dis.readShort();
			
		while (aIIDs!= 0){
			// read the ApplicationInstance ID
			short aIID = dis.readShort();
				
			// read the minimum starting number of actors
			short min = dis.readShort();
				
			// read the number of actors in the application instance
			short actors = dis.readShort(); 
				
			// read the maximum of actors
			short max = dis.readShort();
					
			// init the vector of actors
			Vector actorsVect = new Vector();
					
			while (actors != 0){
				actorsVect.addElement(dis.readUTF());
				actorsVect.addElement(new Short(dis.readShort()));
				actors--;
			}
					
			// form the above datas
			ApplicationInstanceInfos appInfos = new ApplicationInstanceInfos(
					Integer.parseInt(aIID+""),
					min,
					max,
					actorsVect.size()/2,
					actorsVect);
					
			appInstances.addElement(appInfos);
			aIIDs--;
		}
				
		// return the vector of ApplicationInstanceInfos objects
		return appInstances;
	}

	
	/**
	 * Join the specific ApplicationInstance ID.
	 * 
	 * @param sID the Session ID of the user
	 * @param aIID the ApplicationInstance ID choosed
	 * @return the ActorSession ID of the user
	 */
	protected int joinAI(int sID, int aIID) {
		System.out.println("joinAI");
		
		synchronized(monitor){
			String formData = "sID="+sID+"&aIID="+aIID;
	
			try {
			    dis = servletRequest("Join", formData);
				return Integer.parseInt(dis.readShort()+"");
			} 
			catch (Exception e) {
				e.printStackTrace();
			}
			finally { closeAllConnection();}
			
		}
		testMonitorLiberation();
		return 0;
	}

	
	/**
	 * Join randomly an open ApplicationInstance.
	 * 
	 * @param sID the Session ID of the user
	 * @return the ActorSession ID of the user
	 */
	protected int joinAIRnd(int sID) {
		System.out.println("joinAI");
		synchronized(monitor){
			String formData = "sID="+sID+"&aIID=";
	
			try {
			    dis = servletRequest("Join", formData);
				return Integer.parseInt(dis.readShort()+"");
			} 
			catch (Exception e) {
				e.printStackTrace();
			}
			finally { closeAllConnection();}
			
		}
		testMonitorLiberation();
		return 0;
	}
	
	
	/**
	 * Create an public ApplicationInstance.
	 * 
	 * @param sID the Session ID of the user
	 * @param minActors the minimum number of actors required to start
	 * @param maxActors the maximum number of actors 
	 * @return the ActorSession ID of the user
	 */
	protected int createAI(int sID, int minActors, int maxActors) {
		System.out.println("createAI");
		
		synchronized(monitor){
			String formData = "sID="+sID
							+"&minActors="+minActors
							+"&maxActors="+maxActors
							+"&actors=";
	
			try {
			    dis = servletRequest("Create", formData);
				return Integer.parseInt(dis.readShort()+"");
			} 
			catch (Exception e) {
				e.printStackTrace();
			}
			finally { closeAllConnection();}
		}
		testMonitorLiberation();
		return 0;
	}

	
	/**
	 * Create a private Application Instance with the choosed actors.
	 * 
	 * @param sID the Session ID of the user
	 * @param actors the table of ActorID representing the actors accepted in the game session
	 * @param minActors the minimum number of actors required to start
	 * @param maxActors the maximum number of actors
	 * @return the ActorSession ID of the user
	 */
	protected synchronized int createAIPriv(int sID, String[] actors, int minActors, int maxActors) {
		synchronized(monitor){
			String actorsHttp ="";
			for(int i=0; i<actors.length; i++){
				actorsHttp = actorsHttp + "_"+ actors[i];
			}
			
			String formData = "sID="+sID
							+"&minActors="+minActors
							+"&maxActors="+maxActors
							+"&actors="+actorsHttp;
	
			try {
			    dis = servletRequest("Create", formData);
				return Integer.parseInt(dis.readShort()+"");
			} 
			catch (Exception e) {
				e.printStackTrace();
			}
			finally { closeAllConnection();}
			return 0;
		}
	}
    
	
	/**
     * Provide to get the events in user actorSession on the Platform. 
     *
     * @param aSID the ActorSession ID
     * @return a vector of events send by the platform
     */
	protected void getEvents(int aSID){
		System.out.println("getEvents");
		synchronized (monitor){
	        try {
	            dis = servletSendDataRequest("InGame", aSID, new Hashtable());
	            eventsOf(dis);
			} 
			catch (Exception e) {
				e.printStackTrace();
			}
			finally{ 
				closeAllConnection();
			}
		}
		testMonitorLiberation();
    }
	
	/**
	 * In game session, this method provides to the user to change his pseudoname.
	 * Returns a boolean representing the success of the operation. 
	 * 
	 * @param aSID
	 * @return boolean
	 */
    protected String name(int aSID, String name){
    	synchronized(monitor){
		    String formData = "aSID="+aSID+"&name="+name;
			try {
			    dis = servletRequest("Name", formData);
			    String n = dis.readUTF();
			    System.out.println(n);
			    return n;
	        } catch (IOException e) {
	            e.printStackTrace();
	        }
	        finally { closeAllConnection();}
	        return null;
    	}
	}
    
	/**
	 * (Try to) Start the game if the user is the owner of the application instance.
	 * Return 1 if the game has started, else 0.
	 * For manual mode only.
	 * 
	 * @param aSID
	 * @return int
	 */
    protected boolean startAI(int aSID){
    	synchronized(monitor){
	    	String formData = "aSID="+aSID;
		    
			try {
	            dis = servletRequest("StartAI", formData);
	            return dis.readBoolean();
	        } catch (IOException e) {
	            e.printStackTrace();
	        }
	        finally { closeAllConnection();}
	        return false;
    	}
	}
	
	
	/**
	 * (Try to) End the game if the user is the owner of the application instance.
	 * Return 1 if the game has ended, else 0.
	 * For manual mode only.
	 * 
	 * @param aSID
	 */
    protected void endAI(int aSID){
    	synchronized(monitor){
	    	String formData = "aSID="+aSID;
			try {
			    servletRequest("EndAI", formData);
	        } catch (IOException e) {
	            e.printStackTrace();
	        }
	        finally { closeAllConnection();}
    	}
	}
    
    
    /**
	 * Quit the ApplicationInstance joined.
	 * 
	 * @param aSID the ActorSession ID
	 */
    protected void quitAI(int aSID){
    	synchronized(monitor){
	    	String formData = "aSID="+aSID;
			try {
	            servletRequest("QuitAI", formData);
	        } catch (IOException e) {
	            e.printStackTrace();
	        } finally { closeAllConnection();}
    	}
	}
	
	
	 /**
     * Provide to send in game datas to the Platform. 
     *
     * @param aSID the ActorSession ID
     * @param h the hashtable of datas
     * @return a vector of events send by the platform
     */
    protected void sendData(int aSID, Hashtable h){ 
    	synchronized (monitor){
	    	try {
	            dis = servletSendDataRequest("InGame", aSID, h);
	            eventsOf(dis);
			} 
			catch (Exception e) {
				e.printStackTrace();
			}
			finally { closeAllConnection();}
    	}
    }

    static protected HttpConnection connection = null;
	static private DataOutputStream dos = null;
	static private DataInputStream dis = null;
	static private Object monitor = new Object();
	
	/**
	 * Treats the events of the servlet response data input stream.
	 * 
     * @param dis the data input stream
     * @return
	 * @throws IOException
     */
    private void eventsOf(DataInputStream dis) throws IOException {
        byte nb = dis.readByte();
        byte type;
            
        while(nb>0){
            type = dis.readByte(); // get the type of the events
            
            switch (type){
	        	case 1: // JoinEvent
	                onJoinEvent(dis.readShort(),dis.readUTF());
	                break;
	        	case 2: // StartEvent
	        	    onStartEvent(dis.readShort());
	        	    break;
	        	case 3: // EndEvent
	        	    onEndEvent(dis.readShort());
	        	    break;
	        	case 4: // QuitEvent
	        	    onQuitEvent(dis.readShort());
	        	    break;
	        	case 5: 
	        	    try {
                        onDataEvent(dis.readShort(), customTypes.decodeData(dis));
                        break;
                    } catch (IOException e) {
                        e.printStackTrace();
                        break;
                    } catch (Exception e) {
                        e.printStackTrace();
                        break;
                    }
	        	default:  break;
            }
            
            nb--;
        }
    }

	
	/**
	 * Quit the platform.
	 * 
	 * @param sID the Session ID of the user
	 */
    protected void quit(int sID) {
		synchronized(monitor){
	    	String formData = "sID="+sID;
			try {
			  servletRequest("Quit", formData);
	        } catch (IOException e) {
	            e.printStackTrace();
	        } finally { closeAllConnection();}
		}
	}

	
	/**
	 * Send a request to the specified servlet with the http post parameters.
	 * Return the servlet response DataInputStream.
	 * 
	 * @param servletName the name of the servlet
	 * @param formData the HTTP post parameters 
	 * @return the DataInputStream
	 * @throws IOException
	 */
    private DataInputStream servletRequest(String servletName, String formData) throws IOException{
        connection = (HttpConnection) Connector.open(
										GPServletContainerUrl+servletName,
										Connector.READ_WRITE,
										true);
		connection.setRequestMethod(HttpConnection.POST);
		connection.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");
		connection.setRequestProperty("User-Agent",
			"Profile/MIDP-2.0 Configuration/CLDC-1.0");
		connection.setRequestProperty("Content-Language", "fr");
		connection.setRequestProperty("Accept", "application/octet-stream");
			
		byte[] data = formData.getBytes();
		connection.setRequestProperty("Content-Length", Integer.toString(data.length));
			
		dos = connection.openDataOutputStream();
		dos.write(data);
		dos.close();
			
		if (connection.getResponseCode() == HttpConnection.HTTP_OK)
				return connection.openDataInputStream();
		
		else return null;
	}


    /**
     * Send the data in the hashtable to the specified servlet.
     * 
     * @param servletName the name of the servlet
     * @return the DataInputStream of the servlet response
     * @throws IOException
     */
    private DataInputStream servletSendDataRequest(String servletName, int aSID, Hashtable h) throws IOException {
        connection = (HttpConnection) Connector.open(
										GPServletContainerUrl+servletName,
										Connector.READ_WRITE,
										true);
		connection.setRequestMethod(HttpConnection.POST);
		connection.setRequestProperty("Content-Type", "application/octet-stream");
			
		dos = connection.openDataOutputStream();
		try {
		    dos.writeShort(aSID);
		    //System.out.println("Hashtable:"+h.size());
            customTypes.encodeData(h,dos);
            //System.out.println("Encoding ok");
        } catch (Throwable t) {
            t.getMessage();
            t.printStackTrace();
        }
		dos.close();
			
		if (connection.getResponseCode() == HttpConnection.HTTP_OK){
			return connection.openDataInputStream();
		}
		return null;
    }
	
    
    /**
	 * Close all connection streams and the connection.
	 */
	static private void closeAllConnection(){
	    if (dis != null) { try {dis.close();} catch (Exception _e){}}				
		if (dos != null) { try {dos.close();} catch (Exception _e){}}
	    if (connection !=null) {try {connection.close();} catch (Exception e){}}
		connection = null;
	}
    
    
    /** Events management **/
    
    
    /**
     * Method called when a JoinEvent is received from the platform.
     * 
     * @param actorSessionID
     * @param username
     */
    abstract public void onJoinEvent(int actorSessionID, String username);
	
    
    /**
     * Method called when a StartEvent is received from the platform.
     * 
     * @param actorSessionID
     */
    abstract public void onStartEvent(int actorSessionID);

    
    /**
     * Method called when a EndEvent is received from the platform.
     * 
     * @param actorSessionID
     */
    abstract public void onEndEvent(int actorSessionID);
    

    /**
     * Method called when a QuitEvent is received from the platform.
     * 
     * @param actorSessionID
     */
    abstract public void onQuitEvent(int actorSessionID);
    
    
    /**
     * Method called when a DataEvent is received from the platform.
     * 
     * @param actorSessionID
     * @param datas the hashtable of datas
     */
    abstract public void onDataEvent(int actorSessionID, Hashtable datas);

    
    /**
     * Compute delay between open http connection and close http connection ~ ping server
     */
    private long openConnectionTime = 0;
    private long closeConnectionTime = 0;
    
    public int pingServer(){
    	int pingServer = (int)(openConnectionTime - closeConnectionTime);
    	openConnectionTime = 0;
    	closeConnectionTime = 0;
    	return pingServer;
    }
}
