/***
 * GASP: a gaming services platform for mobile multiplayer games.
 * MooDS: a communication protocol used in GASP.
 * GASP Demonstrator: a gasp demonstration game.
 *
 * Copyright (C) 2004 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Franck Caron, Romain Pellerin
 */
/*
 * Created on 2 december 2005
 *
 * @author CARON Franck - CNAM CEDRIC
 * @author PELLERIN Romain - CNAM CEDRIC
 * 
 * 
 * Research project: CNAM-CEDRIC - INT - Filao
 */
//#include parameters.h
////#define DEMO
////#define VERSION_SOLO
////#define FRANCK
////#define LASTPING
////#define RESYNC

//#endinclude
package app2.client;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import app2.CustomTypes;
import app2.NewPlayer;
import app2.PingPlayer;
import app2.UpdatePlayer;

/**
 * @author Maxime Wozny ,Alexandre Garc, Moez Habib Thameur, Franck CARON, Romain PELLERIN
 *
 */
public class NetworkCom extends GASPClient implements Runnable{
	
	private static final boolean DEBUG = true;
	private IHMGasp ihmGasp;
	private Thread comThread;
	public final static long  minimalRequestDeltaTime = 2000;
	
	public NetworkCom(IHMGasp c){
		this.ihmGasp=c;
		comThread = new Thread(this);
		initCustomTypes();
	}

	public static int averagePing = 0;
	
	public void run() {
		try {
			int mode;
			int counter = 0;
			
			while (1 > 0) {
				
				mode=ihmGasp.getMode();
				/**
				 * In the waiting room
				 */
				if (mode==2){
				//#ifdef DEMO
					//# System.out.println("NetworkCom mode 2");
				//#endif DEMO
					try {
				    	getEvents(ihmGasp.aSID);
					}
					catch (Exception e) {
						e.printStackTrace();
					}
					Thread.sleep(2000);
				}
				/**
				 * In game
				 */
				else if (mode==3){
					counter++;
					update();
					
					long systemRequestStartTime = System.currentTimeMillis();
					
					if (sendToNetwork.isEmpty()) {
						getEvents(ihmGasp.aSID);
					} else {
						sendData(ihmGasp.aSID, sendToNetwork);
						releaseAll();
					}
					
					// ping estimate
					int ping = (int)(System.currentTimeMillis() - systemRequestStartTime); 
					averagePing = (ping + (counter-1) * averagePing) / counter;
					
					
					long timeToSleep = minimalRequestDeltaTime - (System.currentTimeMillis() - systemRequestStartTime);
					
				//#ifdef DEMO
					//# if (DEBUG)
						//# System.out.println(" tts:" + timeToSleep);
				//#endif DEMO
					
					if (timeToSleep > 0)
						Thread.sleep(timeToSleep);
				}
				else Thread.sleep(100);
				
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		
	}
	
    private void update() { 
    	try {
    	   	//stage1: examine for a ping
        	Object o1 = ihmGasp.ping();
        	if (o1 != null) addMessageToSend(o1);
        	
        	//stage2: update of position
        	Object o = ihmGasp.update();
        	if (o != null) addMessageToSend(o);

        	//stage3: add all the other elements of the vector in the hashtable
    	    for(Enumeration e = messagesToSend.elements(); e.hasMoreElements();){
    		        addSend(e.nextElement());
    		    }	    
       	    messagesToSend.removeAllElements();
      	
    		}
    	catch (Exception e){    		
    	}
   }
    
    private boolean threadStarted = false;

	public void startComThread(){
		System.out.println("NetworkCom start thread!");
		if (!threadStarted){
			comThread.start();
			threadStarted = true;
		}
	}
	/********************************************/
	/** Listener communication with GASPClient **/
	/********************************************/
	
	public void initCustomTypes() {
		 customTypes = new CustomTypes();
	}

	public void onJoinEvent(int actorSessionID, String username) {
		ihmGasp.onJoinEvent(actorSessionID,username);
	}

	public void onStartEvent(int actorSessionID) {
		ihmGasp.onStartEvent(actorSessionID);
	}

	public void onEndEvent(int actorSessionID) {
		ihmGasp.onEndEvent(actorSessionID);
	}

	public void onQuitEvent(int actorSessionID) {
		ihmGasp.onQuitEvent(actorSessionID);
	}

	public void onDataEvent(int actorSessionID, Hashtable datas) {
		ihmGasp.onDataEvent(actorSessionID,datas);
	}
	
	/**********************************/
	/*** MESSAGE OBJECTS MANAGEMENT ***/
	/**********************************/

    /**
     * message objects pools, to avoid the construction of to many
	 * objets (decrease performance) we use objects pool
     */
	
    private static Vector newPlayerObjects = new Vector();
    private static Vector updatePlayerObjects = new Vector();
    private static Vector pingPlayerObjects = new Vector();
    
    
    /**
	 * Creation of a new event NewPlayer. 
	 * If the vector newPlayerObjects is not empty, the event is put in it,
	 * else the vector is created and the event is put in it.
	 */
	public synchronized static NewPlayer newNewPlayerMessage() {
	    if (DEBUG) System.out.println("newPlayerMessage created");
	    NewPlayer retrieved;
		if (newPlayerObjects.size() > 0) {
			retrieved = (NewPlayer) newPlayerObjects.firstElement();
			newPlayerObjects.removeElement(retrieved);
		} else {
			retrieved = new NewPlayer();
		}
		return retrieved;
	}
	
	/**
	 * Creation of a new event UpdatePlayer. 
	 * If the vector upDatePlayerObjects is not empty, the event is put in it,
	 * else the vector is created and the event is put in it.
	 * @param : The element's ID to be deleted
	 */
	public synchronized static UpdatePlayer newUpdatePlayerMessage(byte type) {
	    if (DEBUG) System.out.println("newGASPMessage type>"+type);
	    UpdatePlayer retrieved;
		if (updatePlayerObjects.size() > 0) {
			retrieved = (UpdatePlayer) updatePlayerObjects.firstElement();
			updatePlayerObjects.removeElement(retrieved);
		} else {
			retrieved = new UpdatePlayer();
		}
		return retrieved;
	}
	/**
	 * Creation of a new event PingPlayer. 
	 * If the vector pingPlayerObjects is not empty, the event is put in it,
	 * else the vector is created and the event is put in it.
	 */
	public synchronized static PingPlayer newPingPlayerMessage(byte type) {
	    if (DEBUG) System.out.println("PingPlayer Message created"+type);
	    PingPlayer retrieved;
		if (pingPlayerObjects.size() > 0) {
			retrieved = (PingPlayer) pingPlayerObjects.firstElement();
			pingPlayerObjects.removeElement(retrieved);
		} else {
			retrieved = new PingPlayer();
		}
		return retrieved;
	}	
	
	/**
	 * release the object which is put in parameter. 
	 * Put the object in the associated vector.
	 * @param : x object to release
	 */
	public static synchronized  void release(Object x) {
		if (x instanceof NewPlayer)
		    newPlayerObjects.addElement(x);
		else if (x instanceof UpdatePlayer)
			updatePlayerObjects.addElement(x);
		else if (x instanceof PingPlayer)
			pingPlayerObjects.addElement(x);
	}

	/**
	 * release all the elements of the hashtable and get it empty
	 * 
	 */
	private synchronized static void releaseAll(){
		Enumeration enume;
		for (enume = sendToNetwork.elements(); enume.hasMoreElements(); ) {
			Object current = (Object) enume.nextElement();
			/* TODO if (!(current.equals(updateHero))){
			    if (DEBUG) System.out.println("release object type>"+current.getType()+" MID>"+ current.getId());
			  
				current.raz();*/
			    release(current);
			
		}
		nbSend = new Integer(0);
		sendToNetwork.clear();
	}
    
     //TODO (possibly types's factorization) ie, private final int TYPE_NEWPLAYER = 6; ...
    
	/***********************/
	/*** SEND MANAGEMENT ***/
	/***********************/
	
    // hastable management (sendData method of GASPClient)
    private static Hashtable sendToNetwork = new Hashtable();
    private static Integer nbSend = new Integer(0);
    
	/**
	 * Add an object in the hashtable send
	 * @param : x object to be added
	 */
	protected synchronized static void addSend(Object x){
	    /* TODO if (x.equals(updateHero))
	        if (send.contains(updateHero)) return;
	   */  
	//#ifdef DEMO
		//# System.out.println("Put msg object in Hashtable: "+x.toString());
	//#endif DEMO
		sendToNetwork.put(nbSend.toString(),x);
	    nbSend = new Integer(nbSend.intValue()+1);
	}
    
	/*********************************************/
    /*** INCOMING MESSAGE TO SEND FROM IHMGasp ***/
	/*********************************************/
	
	/**
	 *  In game: testing this vector which contains the data to be sent (position...)
	 *  
	 */
	private static Vector messagesToSend = new Vector();
	
	public void addMessageToSend(Object o) {
		messagesToSend.addElement(o);
	}	
	
}
