/***
 * GASP: a gaming services platform for mobile multiplayer games.
 * MooDS: a communication protocol used in GASP.
 * GASP Demonstrator: a gasp demonstration game.
 *
 * Copyright (C) 2004 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Franck Caron, Romain Pellerin
 */
/*
 * Created on 2 december 2005
 *
 * @author CARON Franck - CNAM CEDRIC
 * @author PELLERIN Romain - CNAM CEDRIC
 * @author DELPIANO Fabien - Filao
 * 
 * 
 * Research project: CNAM-CEDRIC - INT - Filao
 */
//#include parameters.h
////#define DEMO
////#define VERSION_SOLO
////#define FRANCK
////#define LASTPING
////#define RESYNC

//#endinclude
package app2.client.game;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;

public abstract class Actor {

	private static final int[] WALKING_IMAGES = { 3, 4 };
	private static final int[] WALKING_TIMING = { 200, 200 };

	private int totalAnimationLoopTime;
	protected GameCanvas game;

	protected int[] images;
	protected int[] timing;
	
	protected String username;

	public Actor(GameCanvas game) {
		this.game = game;

		images = WALKING_IMAGES;
		timing = WALKING_TIMING;

		for (int i = 0; i < timing.length; i++) {
			totalAnimationLoopTime += timing[i];
		}

		initialize();
	}
	
	/***********************/
	/*** Initialization ***/
	/**********************/

	public abstract void initialize();


	/**************************/
	/*** Position and Speed ***/
	/**************************/

	protected int x, y;
	private int speed_x, speed_y;

	public int getX() {
		return x;
	}

	public int getY() {
		return y;
	}
	
	public int getSpeedX() {
		return speed_x;
	}

	public int getSpeedY() {
		return speed_y;
	}

	public void moveX(int dx, long delta_time) {
		x += dx * delta_time;
		if (x < 0) x = 0;
		else if (x > game.xMax) x = game.xMax;
	}

	public void moveY(int dy, long delta_time) {
		y += dy * delta_time;
		if (y < 0) y = 0;
		else if (y > game.yMax) y = game.yMax;
	}

	public void setSpeed(int speed_x, int speed_y) {
		this.speed_x = speed_x;
		this.speed_y = speed_y;
	}

	/**************************/
	/*** In game management ***/
	/**************************/

	private boolean in_game = false;

	public void addedNotify() {
		in_game = true;
	}

	public void removedNotify() {
		in_game = false;
	}

	public boolean isInGame() {
		return in_game;
	}

	/****************/
	/*** Actions ***/
	/***************/
	
	public void executeActions(long time, long delta_time) {
		moveX(speed_x, delta_time);
		moveY(speed_y, delta_time);
	}

	/*********************/
	/****** Graphic ******/
	/*********************/

	/**
	 * @param g
	 * @param gameTime
	 * @param deltaTime
	 */
	public void paint(Graphics g, long gameTime, long deltaTime) {

		/**
		 * Animation
		 */
		int image_index = 0;
		int time = (int) (gameTime % totalAnimationLoopTime);
		int sum = timing[0];
		for (; image_index < timing.length; image_index++) {
			if (time < sum)
				break;
			sum += timing[image_index];
		}

		Image img = ResourcesManager.getImage(images[image_index]);
		if (img != null)
			g.drawImage(img, game.convertToScreenX(getX()), game.convertToScreenY(getY()), Graphics.HCENTER | Graphics.VCENTER);

    }

	
	/*****************************/
	/*** Collisions management ***/
	/*****************************/

	/**
	 * This part is to manage collision between actor and/or other objects. It's not include in this demostrator.
	 */
//#ifdef DEMO
	//# 
	//# //Collision management -----------------------------------	
	//# 
	//# protected int width = 200000, height = 200000;
	//# 
	//# public boolean isCollidableWith(Actor actor) {
		//# return false;
	//# }
	//# public boolean isMainCollider() {
		//# return false;
	//# }
	//# public boolean isNotifiedForCollision() {
		//# return false;
	//# }
	//# 
	//# public void addCollision(Actor _actor, long time, long delta_time) {}
	//# 
	//# public boolean isCollidingWith(Actor actor) {
		//# 
		//# int x_start = x - width / 2, x_end = x + width / 2;
		//# int y_start = y - height / 2, y_end = y + height / 2;
		//# int other_x_start = actor.x - actor.width / 2, other_x_end = actor.x + actor.width / 2;
		//# int other_y_start = actor.y - actor.height / 2, other_y_end = actor.y + actor.height / 2;
		//# 
		//# boolean overlap_on_x = segmentsOverlap(x_start , x_end, other_x_start , other_x_end);
		//# boolean overlap_on_y = segmentsOverlap(y_start , y_end, other_y_start , other_y_end);
	//# 
		//# return overlap_on_x && overlap_on_y;
	//# }
	//# 
	//# private boolean segmentsOverlap(int one_start, int one_end, int two_start, int two_end) {
	//# 
		//# if (inBounds(one_start, two_start, two_end))
			//# return true;
		//# 
		//# if (inBounds(one_end, two_start, two_end))
			//# return true;
	//# 
		//# if (two_start < one_start && two_end > one_end) return true;
		//# 
		//# return false;
	//# 
	//# }
	//# 
	//# static public boolean inBounds(int n, int min, int max) {
		//# return n >= min && n <= max;
	//# }
//#endif DEMO
}
