/***
 * GASP: a gaming services platform for mobile multiplayer games.
 * MooDS: a communication protocol used in GASP.
 * GASP Demonstrator: a gasp demonstration game.
 *
 * Copyright (C) 2004 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Franck Caron, Romain Pellerin
 */
/*
 * Created on 2 december 2005
 *
 * @author CARON Franck - CNAM CEDRIC
 * @author PELLERIN Romain - CNAM CEDRIC
 * @author DELPIANO Fabien - Filao
 * 
 * 
 * Research project: CNAM-CEDRIC - INT - Filao
 */
//#include parameters.h
////#define DEMO
////#define VERSION_SOLO
////#define FRANCK
////#define LASTPING
////#define RESYNC

//#endinclude
package app2.client.game;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import app2.client.*;
import app2.*;

import javax.microedition.lcdui.Canvas;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Graphics;

public abstract class GameCanvas extends Canvas implements Runnable, CommandListener {


	private App midlet;
	
	private Command pauseCommand, endsessionCommand, quitsessionCommand, quitCommand;
	private Thread mainloop;

	private boolean paused;
	private boolean isOwner;
	
	public static final int ZOOM_FACTOR = 10000;
	public final int xMax = 2200000;
	public final int yMax = 2400000;
	public static final int NB_OF_PNGS = 8;
	public static final boolean DEBUG = true;
	

	protected int mode  ;
	/**
	 * @param midlet
	 */
	public GameCanvas(App midlet, IHMGasp ihmGasp, int _time) {

		this.midlet = midlet;
		this.ihmGasp = ihmGasp;
		
		setCommandListener(this);

		mainloop = new Thread(this);
		mainloop.start();

		networkDeltaTime = _time;
		
	/************************/
	/*** Canvas commands ****/
	/************************/
		
		pauseCommand = new Command("Pause", Command.SCREEN, 1);
		addCommand(pauseCommand);
		
		isOwner = ihmGasp.isOwner;
		
		if (isOwner) {
			endsessionCommand = new Command("EndSession", Command.SCREEN, 2);
			addCommand(endsessionCommand);
		}
		
		quitsessionCommand = new Command("QuitSession", Command.SCREEN, 2);
		addCommand(quitsessionCommand);
		
		quitCommand = new Command("Quit", Command.SCREEN, 2);
		addCommand(quitCommand);
		
	}
	
	public abstract void initialize();

	/**************************/
	/*** Canvas management ****/
	/**************************/

	
	public void commandAction(Command c, Displayable d) {

		if (c == pauseCommand) {
			if (!paused) {
				pause();
			} else {
				restart();
			}
		} 
		 else if (c == endsessionCommand) {
			 pause();
			 ihmGasp.endAI();
			 ihmGasp.mode = 2;
			 ihmGasp.currentScreen();
		} else if (c == quitsessionCommand) {
			pause();
			ihmGasp.quitAI();
			ihmGasp.mode = 1;
			ihmGasp.currentScreen();
		} else if (c == quitCommand) {
			quit();
			ihmGasp.quit();
			ihmGasp.mode = 0;
			ihmGasp.currentScreen();
		}
		
		}

	/************************/
	/**** Key management ****/
	/************************/

	private int[] downKeys = new int[3];

	private void iniKeys() {
		for (int i = 0 ; i < downKeys.length ; i++) {
			downKeys[i] = Integer.MAX_VALUE;
		}
	}

	public void keyPressed(int keyCode) {
		int gameAction = getGameAction(keyCode);
		for (int i = 0 ; i < downKeys.length ; i++) {
			if (downKeys[i] == gameAction) return;
			if (downKeys[i] == Integer.MAX_VALUE) {
				downKeys[i] = gameAction;
				return;
			} 
		}
	}
	
	protected void keyReleased(int keyCode) {
		int gameAction = getGameAction(keyCode);
		for (int i = 0 ; i < downKeys.length ; i++) {
			if (downKeys[i] == gameAction) {
				downKeys[i] = Integer.MAX_VALUE;
				return;
			} 
		}
	}
	
	public boolean isKeyPressed(int gameAction) {
		for (int i = 0 ; i < downKeys.length ; i++) {
			if (downKeys[i] == gameAction) return true;
		}
		return false;
	}

	/**************************/
	/*** Graphic Rendering ****/
	/**************************/

	protected void paint(Graphics g) {
		g.setColor(0xFFFFCC);
		g.fillRect(0 , 0 , getWidth() , getHeight());
		
		for (Enumeration enume = actors_in_game.elements() ; enume.hasMoreElements();) {
			Actor current = (Actor)enume.nextElement();
			current.paint(g, gameTime, deltaTime);		
		}
	//#ifdef DEMO	
		//# g.setColor(0xFFCC33);
		//# g.drawString("Score : " + score , 2, 2, Graphics.TOP | Graphics.LEFT);
		//# g.setColor(0xFF0000);
		//# g.drawString("Lives : " + lives , 2, getHeight() - 2, Graphics.BOTTOM | Graphics.LEFT);
	//#endif DEMO
	}

	/***********************************/
	/*** Game life cycle management ****/
	/**********************************/
	
	public void newGame() {
		iniKeys();	
		reInitializeForNewGame();
		initializeNewGame();
		restart();
	}
	
	protected abstract void initializeNewGame();
	
	private void reInitializeForNewGame() {
		
		fireActors();
		fireAllActorsInGame();
		actors_to_be_added.removeAllElements();
	//#ifdef DEMO	
		//# lives = 3;
		//# score = 0;
	//#endif DEMO	
	}

	public synchronized void pause() {
		paused = true;
	}
	
	public synchronized void restart() {
		paused = false;
		notifyAll();
	}
	
	private synchronized void quit() {
		paused = true;
		
		//#ifdef VERSION_SOLO
		//# midlet.backToIntro();
		//#endif VERSION_SOLO
	}
	
	public void stop() {}

	
	/******************/
	/*** Main loop ****/
	/******************/

	long gameStartTime = 0, gameTime = 0, deltaTime = 0;
	int networkDeltaTime = 0;

	private void initTime() {
		gameStartTime = System.currentTimeMillis();
		deltaTime = 0;
		gameTime = 0;
	}
	
	public boolean canUpdate = false;
	
	private void updateTime() {
		long now = System.currentTimeMillis() - gameStartTime;
		deltaTime = now - gameTime;
		gameTime = now;
	}

	public void run() {

		while (true) {
			try{
			updateTime();
			checkMessagesReceived();
			manageCamera();
			hireActors();
			fireActors();
		//#ifdef DEMO
			//# computeCollisions(gameTime , deltaTime);
		//#endif DEMO
			executeAllActorsBehaviours(gameTime , deltaTime);
			repaint();
			}
			catch (Exception e){
			}
			synchronized (this) {
				if (paused) {
					long interruptionTime = System.currentTimeMillis();
					try {
						wait();
					} catch (InterruptedException e) {
						continue;
					}
					gameStartTime += System.currentTimeMillis() - interruptionTime;
				}
			}

		}

	}
	
	/**
	 * @param gameTime
	 * @param deltaTime
	 */
	private void executeAllActorsBehaviours(long gameTime, long deltaTime) {
		for (Enumeration enume = actors_in_game.elements() ; enume.hasMoreElements();) {
			Actor current = (Actor)enume.nextElement();
			current.executeActions(gameTime, deltaTime);
		}
	}

	
	/**************************/
	/*** Camera management ****/
	/**************************/

	
	protected void manageCamera() {}

	/*****************************/
	/*** Collisions management ***/
	/*****************************/
	
	/**
	 * This part is to manage collision between actor and/or other objects. It's not include in this demostrator.
	 */

//#ifdef DEMO
	//# private void computeCollisions(long time, long delta_time) {
		//# for (Enumeration maincolliders = actors_in_game.elements() ; maincolliders.hasMoreElements();) {
			//# Actor currentmain = (Actor)maincolliders.nextElement();
			//# if (!currentmain.isMainCollider()) continue;
			//# for (Enumeration secondarycolliders = actors_in_game.elements() ; secondarycolliders.hasMoreElements();) {
				//# Actor currentsecondary = (Actor)secondarycolliders.nextElement();
				//# if (currentmain == currentsecondary) continue;
				//# if (currentmain.isCollidableWith(currentsecondary)) {
					//# boolean is_colliding = currentmain.isCollidingWith(currentsecondary);
					//# if (is_colliding) {
						//# if (currentmain.isNotifiedForCollision()) currentmain.addCollision(currentsecondary, time, delta_time);
						//# if (currentsecondary.isNotifiedForCollision()) currentsecondary.addCollision(currentmain, time, delta_time);
					//# }
				//# }
			//# }
		//# }
	//# }
//#endif DEMO	
	

	/*************************/
	/*** Actors management ***/
	/*************************/
	
	private Vector actors_in_game = new Vector(),
				   actors_to_be_added = new Vector(),
				   actors_to_be_removed = new Vector();
	
	public boolean addActor(Actor actor) {
		if (actor.isInGame()) return false;
		actors_to_be_added.addElement(actor);
		return true;
	}
	
	public void removeActor(Actor actor) {
		actors_to_be_removed.addElement(actor);
	}
	
	private void hireActors() {
		for (Enumeration enume = actors_to_be_added.elements() ; enume.hasMoreElements();) {
			Actor added = (Actor)enume.nextElement();
			actors_in_game.addElement(added);
			added.addedNotify();
			System.out.println("Add: "+added);
		}
		actors_to_be_added.removeAllElements();
	}
	
	private void fireActors() {
		for (Enumeration enume = actors_to_be_removed.elements() ; enume.hasMoreElements();) {
			Actor removed = (Actor)enume.nextElement();
			boolean ok = actors_in_game.removeElement(removed);
			if (ok) removed.removedNotify();
		}
		actors_to_be_removed.removeAllElements();
	}
	
	private void fireAllActorsInGame() {
		for (Enumeration enume = actors_in_game.elements() ; enume.hasMoreElements();) {
			Actor removed = (Actor)enume.nextElement();
			boolean ok = actors_in_game.removeElement(removed);
			if (ok) removed.removedNotify();
		}
	}
	
	/**
	 * method to get a Multiplayer Actor. mid define this special Actor
	 * @param mid
	 * @return
	 */
	public synchronized MultiplayerActor getMultiplayerActor(int mid){
		for(int i = 0 ; i < actors_to_be_added.size(); i++){
			Actor a = (Actor) actors_to_be_added.elementAt(i);
			if (a instanceof MultiplayerActor){
				MultiplayerActor ma = (MultiplayerActor) a;
				if (ma.multiplayerID == mid) return ma;
			}
		}
		for(int i = 0 ; i < actors_in_game.size(); i++){
			Actor a = (Actor) actors_in_game.elementAt(i);
			if (a instanceof MultiplayerActor){
				MultiplayerActor ma = (MultiplayerActor) a;
				if (ma.multiplayerID == mid) return ma;
			}
		}
		return null;
	}
	
	public int nbMultiplayerActors = 0;
	
	/******************************/
	/*** Coordinates conversion ***/
	/******************************/

	private int view_center_x = 1000000, view_center_y = 1000000;

	
	public int convertToScreenX(int x) {
		return getWidth() / 2 + (int)((x - view_center_x) / ZOOM_FACTOR
				);
	}
	
	public int convertToScreenY(int y) {
		return getHeight() / 2 + (int)((y - view_center_y) / ZOOM_FACTOR
				);
	}
	
	public void centerOnActorX(Actor actor) {
		view_center_x = actor.getX();
	}
	
	public void centerOnActorY(Actor actor) {
		view_center_y = actor.getY();
	}
	
	 public IHMGasp ihmGasp;
	 
//#ifdef DEMO
	//# /**
	 //# * Score and lives ------------------------------------------
	 //# */
	//# 
	//# private int lives;
	//# private int score;
	 //# 
	//# public void looseLife() {
		//# //lives--;
		//# if (lives <= 0) quit();
	//# }
	//# public void earnPoints(int points) {
		//# this.score += points;
	//# }
//#endif DEMO

	/***************************/
	/*** Messages management ***/
	/***************************/

	private static Vector messagesReceived = new Vector();

	/**
	 * When an actor quits a game, it (and its multiplayerActor) must be removed
	 * @param actorSessionID
	 */
	public void onQuitEvent(int actorSessionID) {
		try{
			// destruction of the actor which have this actorSessionID
			MultiplayerActor act = getMultiplayerActor(actorSessionID);
			removeActor(act);
		}
		catch (Exception e){
		}
		
	}
		
	/**
	 * for a specific Actor, add messages to be treated 
	 * @param actorSessionID
	 * @param datas
	 */
	public void onDataEvent(int actorSessionID, Hashtable datas) {
		int nbSend = datas.size();
		for (int i = 0; i < nbSend; i++) {
			Object o = (Object) datas.get("" + i);
			messagesReceived.addElement(o);
		}
	}
	
	/**
	 * treat received messages from NetworkCom after HTTP request
	 *
	 */
	private void checkMessagesReceived() {	
		for (int i = 0; i < messagesReceived.size(); i++ ){
			Object o = messagesReceived.elementAt(i);
			treatReceivedMessage(o);
		}
		messagesReceived.removeAllElements();
	}
	
	public abstract void treatReceivedMessage(Object o);
	public abstract Object cyclicUpdateMessagesToSend();
	public abstract Object cyclicPingPlayerMessagesToSend();

}
