/***
 * GASP: a gaming services platform for mobile multiplayer games.
 * MooDS: a communication protocol used in GASP.
 * GASP Demonstrator: a gasp demonstration game.
 *
 * Copyright (C) 2004 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Franck Caron, Romain Pellerin
 */
/*
 * Created on 2 december 2005
 *
 * @author CARON Franck - CNAM CEDRIC
 * @author PELLERIN Romain - CNAM CEDRIC
 * @author DELPIANO Fabien - Filao
 * 
 * 
 * Research project: CNAM-CEDRIC - INT - Filao
 */
//#include parameters.h
////#define DEMO
////#define VERSION_SOLO
////#define FRANCK
////#define LASTPING
////#define RESYNC

//#endinclude
package app2.client.game;

import java.util.Vector;

import app2.NewPlayer;
import app2.PingPlayer;
import app2.UpdatePlayer;
import app2.client.App;
import app2.client.IHMGasp;
import app2.client.NetworkCom;


public class SimpleGameCanvas extends GameCanvas {

	private HeroeActor heroe;
	//#ifdef DEMO
	//# private BallSpitter ballSpitter;
	//#endif DEMO
	
	/**
	 * @param midlet
	 */
	public SimpleGameCanvas(App midlet, IHMGasp ihmGasp, int _time) {
		super(midlet, ihmGasp,_time);
	}

	/*********************************/
	/*** SimpleGame initialization ***/
	/*********************************/

	protected void initializeNewGame() {
		
		if (heroe == null) {
			heroe = new HeroeActor(this);
			heroe.username = ihmGasp.pseudoname;
			//#ifdef DEMO
			//# ballSpitter = new BallSpitter(this);
			//#endif DEMO
		} else {
			heroe.initialize();
			//#ifdef DEMO
			//# ballSpitter.initialize();
			//#endif DEMO
		}
		
		//#ifdef DEMO
		//# ballSpitter.initializeSpitter(1800000, 500000);
		//#endif DEMO
		
		addActor(heroe);
		//#ifdef DEMO
		//# addActor(ballSpitter);
		//#endif DEMO
		
		sendNewPlayer();
		canUpdate = true;
	}
	/**
	 * This method sends a newPlayer message. It's called by  
	 * initializeNewGame
	 */
	private void sendNewPlayer() {
		NewPlayer np = NetworkCom.newNewPlayerMessage();
		np.setId(ihmGasp.aSID);
		np.setX(heroe.getX());
		np.setY(heroe.getY());
		np.setUsername(heroe.username);
		np.draw();
		System.out.println("Push NewPlayer msg to NetworkCom");
		ihmGasp.sendMessage(np);
		System.out.println("Msg correctly pushed");
	}
	
	/**
	 * this method preloads all the PNG
	 */
	public void initialize() {
				for (int i = 1 ; i <=
				//#ifdef VERSION_SOLO
				//# App.NB_OF_PNGS
				//#else
				//#endif VERSION_SOLO
				GameCanvas.NB_OF_PNGS
				; i++) {
			try{
			ResourcesManager.preloadResource(i , ResourcesManager.PNG);
				}
			catch (Exception e){
				System.out.println(i);
				}
			}
		ResourcesManager.loadEverything();
	}
	
	protected void manageCamera() {
		//if (heroe != null) centerOnActorX(heroe);
	}

	/***************************/
	/*** Messages management ***/
	/***************************/
	
	/**
	 * This method treats each type of message:
	 * UpdatePlayer, NewPlayer, PingPlayer 
	 */
	public void treatReceivedMessage(Object o) {
		if (o instanceof UpdatePlayer){
			try{
				UpdatePlayer updp = (UpdatePlayer) o;
				MultiplayerActor ma = getMultiplayerActor(updp.getId());
					
				if (ma!=null){
					ma.setWantedPosition(updp.getX(), updp.getY());
					ma.setWantedPositionSpeeds(updp.getSpeedX(), updp.getSpeedY());
					ma.isPositionReached = false;
				}else{
				/*	System.out.println("Player unknown");*/
					}
			}catch(NullPointerException ne){
				/*System.out.println("here :p");*/
			}
		}
		else if (o instanceof NewPlayer) {
			try{
				NewPlayer np = (NewPlayer) o;
				np.draw();
				MultiplayerActor ma = new MultiplayerActor(this,np.getId());
				nbMultiplayerActors++;
				ma.username = np.getUsername();
				System.out.println("Player Name:"+np.getUsername());
				ma.initialize();
				addActor(ma);	
			}catch(NullPointerException ne){
				System.out.println("here");
			}
		}
		else if (o instanceof PingPlayer){
			try{
				PingPlayer pp = (PingPlayer) o;
			//#ifdef DEMO
				//# if (DEBUG) System.out.println("PingPlayer received ("+pp.getId()+","+pp.getPingNum()+")");
			//#endif DEMO
				MultiplayerActor ma = getMultiplayerActor(pp.getId());
				ma.updatePing(pp.getPingNum(), pp.getSenderAveragePing(), pp.getDesynchronizationWithSender());
			}catch(NullPointerException ne){
				if (DEBUG) System.out.println("Error from a treatment of ping: null");
			}
		}
	}
	
	
	private static UpdatePlayer updp = new UpdatePlayer();
	private static PingPlayer pp = new PingPlayer();
	public static long lastPingUpdateTime = 0;
	
	/**
	 * This object is used to send in a cyclic way the Update message 
	 */
	public Object cyclicUpdateMessagesToSend() {
		if (canUpdate){
			updp.setId(ihmGasp.aSID);
			updp.setX(heroe.getX());
			updp.setY(heroe.getY());
			updp.setSpeedX(heroe.getSpeedX());
			updp.setSpeedY(heroe.getSpeedY());
			
			canUpdate = false;
			return updp;
		}
		
		else return null;
	}
	
	/**
	 * This object is used to send in a cyclic way the PingPlayer message 
	 */
	
	public Object cyclicPingPlayerMessagesToSend(){
		pp.setId(ihmGasp.aSID);
		pp.setPingNum(pp.getPingNum()+1);
	//#ifdef DEMO
		//# System.out.println("PingPlayer sent num: "+pp.getPingNum());
	//#endif DEMO
		return pp;
		
	}
	
	
	
	/***************** kezako dbut ************/
	
	public void checkUpdatePing(){
		System.out.println("Nb ma: "+nbMultiplayerActors);
		for(int i=1; i<=nbMultiplayerActors; i++){
			System.out.println("update ping ma n: "+i);
			try{
				MultiplayerActor ma = getMultiplayerActor(i);
				System.out.println("retrieve ma ok");
				if (!ma.pingUpdated) {
					System.out.println("=> ping missed");
					ma.lastPingTime += NetworkCom.minimalRequestDeltaTime;
				}
			}
			catch (Exception e){}
		}
	}
	
	public void reinitPingUpdatedBooleans(){
		System.out.println("reinit pingUpdated");
		for(int i=1; i<=nbMultiplayerActors; i++){
			try{
				MultiplayerActor ma = getMultiplayerActor(i);
				ma.pingUpdated = false;
				System.out.println("pingUpdated reinitiated for ma n"+i);
			}
			catch (Exception e){}
		}
	}
	/***************** kezako fin ************/

}
